"""Tests for the GraphQL interface for gbp-fl"""

from dataclasses import replace
from typing import Any
from unittest.mock import Mock, patch

from django.test.client import Client
from gentoo_build_publisher.records import BuildRecord
from unittest_fixtures import (
    FixtureContext,
    FixtureOptions,
    Fixtures,
    TestCase,
    depends,
    requires,
)

from gbp_fl import graphql as gql
from gbp_fl.records import Repo
from gbp_fl.types import BinPkg, Build

# pylint: disable=missing-docstring


@depends("settings")
def repo_fixture(_options: FixtureOptions, fixtures: Fixtures) -> FixtureContext[Repo]:
    repo: Repo = Repo.from_settings(fixtures.settings)

    with patch("gbp_fl.graphql.Repo.from_settings", return_value=repo):
        yield repo


@requires(repo_fixture, "bulk_content_files")
class FileListSearchTests(TestCase):
    def test_search_without_machine(self) -> None:
        f = self.fixtures
        repo = f.repo
        repo.files.bulk_save(f.bulk_content_files)

        query = """
          query filesStartingWithBa {
            flSearch(key: "ba*") { path binpkg { cpvb } }
          }
        """
        result = graphql(query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 4)

    def test_search_without_machine_no_match(self) -> None:
        query = 'query { flSearch(key: "python") { path binpkg { cpvb } } }'

        result = graphql(query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 0)

    def test_search_with_machine(self) -> None:
        f = self.fixtures
        repo = f.repo
        repo.files.bulk_save(f.bulk_content_files)
        query = """
          query {
            flSearch(key: "ba*", machine: "polaris") {
              path binpkg { cpvb repo url }
            }
          }
        """
        result = graphql(query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(len(result["data"]["flSearch"]), 3)


@requires(repo_fixture, "bulk_content_files")
class ResolveQueryCountTests(TestCase):
    query = "query totalFileCount { flCount }"

    query_with_machine = """
      query totalFileCountMachine($machine: String!) {
        flCount(machine: $machine)
      }
    """
    query_with_build = """
      query totalFileCountMachine($machine: String!, $buildId: String!) {
        flCount(machine: $machine, buildId: $buildId)
      }
    """

    def test(self) -> None:
        f = self.fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(self.query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 6)

    def test_with_no_content_files(self) -> None:
        result = graphql(self.query)

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 0)

    def test_with_machine(self) -> None:
        f = self.fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(self.query_with_machine, machine="lighthouse")

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 2)

    def test_with_build(self) -> None:
        f = self.fixtures
        repo = f.repo

        repo.files.bulk_save(f.bulk_content_files)
        result = graphql(self.query_with_build, machine="polaris", buildId="26")

        self.assertTrue("errors" not in result, result.get("errors"))
        self.assertEqual(result["data"]["flCount"], 3)


@requires(repo_fixture, "publisher", "build_record", "bulk_content_files")
class ResolveBinPkgBuildTests(TestCase):
    def test(self) -> None:
        f = self.fixtures
        publisher = f.publisher
        build_record: BuildRecord = replace(f.build_record, submitted=f.now)
        build = Build(machine=build_record.machine, build_id=build_record.build_id)
        binpkg = BinPkg(
            build=build,
            cpvb="dev-language/python-3.13.1-3",
            repo="gentoo",
            build_time=self.fixtures.now,
        )
        publisher.repo.build_records.save(build_record)
        result = gql.resolve_binpkg_build(binpkg, Mock())

        self.assertEqual(result, build_record)


def graphql(query: str, **variables: Any) -> Any:
    """Execute GraphQL query on the Django test client

    Return parse JSON response
    """
    post_data = {"query": query, "variables": variables or None}
    client = Client(raise_request_exception=True)
    response = client.post("/graphql", post_data, content_type="application/json")

    return response.json()



"""unittest fixtures"""

# pylint: disable=missing-docstring

import datetime as dt
import io
import os
import tempfile
from contextlib import ExitStack
from pathlib import PurePath as Path
from typing import Any, Mapping, Sequence
from unittest import mock

import rich.console
from django.test.client import Client
from gbpcli.gbp import GBP
from gbpcli.theme import DEFAULT_THEME
from gbpcli.types import Console
from gentoo_build_publisher import publisher as publisher_obj
from gentoo_build_publisher import types as gbp
from gentoo_build_publisher import worker as gbp_worker
from gentoo_build_publisher.build_publisher import BuildPublisher
from gentoo_build_publisher.records import BuildRecord
from gentoo_build_publisher.settings import Settings as GBPSettings
from requests import PreparedRequest, Response
from requests.adapters import BaseAdapter
from requests.structures import CaseInsensitiveDict
from rich.theme import Theme
from unittest_fixtures import FixtureContext, FixtureOptions, Fixtures, depends

from gbp_fl.records import Repo
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, ContentFile, Package

COUNTER = 0


################
# gbp-fl stuff #
################
@depends("tmpdir", "environ")
def settings(_options: FixtureOptions, _fixtures: Fixtures) -> Settings:
    return Settings.from_environ()


@depends("settings")
def repo(options: FixtureOptions, fixtures: Fixtures) -> FixtureContext[Repo]:
    where: str = options.get("repo", {}).get("where", "gbp_fl.records.Repo")
    repo_: Repo = Repo.from_settings(fixtures.settings)

    with mock.patch(f"{where}.from_settings", return_value=repo_):
        yield repo_


@depends("binpkg", "now")
def content_file(options: FixtureOptions, fixtures: Fixtures) -> ContentFile:
    args = get_options(
        options.get("content_file", {}),
        binpkg=fixtures.binpkg,
        path=Path("/bin/bash"),
        timestamp=fixtures.now,
        size=870400,
    )
    return ContentFile(**args)


@depends("now")
def bulk_content_files(
    options: FixtureOptions, fixtures: Fixtures
) -> list[ContentFile]:
    content_files: list[ContentFile] = []
    cf_defs: str = options.get("bulk_content_files", DEFAULT_CONTENTS).strip()
    for cf_def in cf_defs.split("\n"):
        cf_def = cf_def.strip()

        if not cf_def:
            continue

        parts = cf_def.split()
        machine, build_id, cpvb, path = parts[:4]

        try:
            repo_ = parts[4]
        except IndexError:
            repo_ = "gentoo"

        bld = Build(machine=machine, build_id=build_id)
        pkg = BinPkg(build=bld, cpvb=cpvb, build_time=fixtures.now, repo=repo_)
        content_files.append(
            ContentFile(
                binpkg=pkg, path=Path(path), timestamp=fixtures.now, size=850648
            )
        )

    return content_files


@depends("now")
def bulk_packages(options: FixtureOptions, fixtures: Fixtures) -> list[Package]:
    packages: list[Package] = []

    for p_def in options.get("bulk_packages", "").strip().split("\n"):
        p_def = p_def.strip()

        if not p_def:
            continue

        parts = p_def.split()
        cpv = parts[0]

        build_id = seq_get(parts, 1, 1)

        try:
            build_time = int(parts[2])
        except IndexError:
            build_time = fixtures.now.timestamp()

        # crude parsing, but good enough for now
        c, pv = cpv.split("/", 1)
        p, v = pv.rsplit("-", 1)
        if v.startswith("r"):
            p, rest = p.rsplit("-", 1)
            v = f"{rest}-{v}"
        path = f"{c}/{p}/{pv}-{build_id}.gpkg.tar"

        package = Package(
            cpv=cpv,
            repo=seq_get(parts, 3, "gentoo"),
            build_id=build_id,
            build_time=build_time,
            path=path,
        )
        packages.append(package)
    return packages


@depends("build", "now")
def binpkg(options: FixtureOptions, fixtures: Fixtures) -> BinPkg:
    args = get_options(
        options.get("package", {}),
        build=fixtures.build,
        cpvb="app-shells/bash-5.2_p37-3",
        build_time=fixtures.now,
        repo=options.get("repo", "gentoo"),
    )
    return BinPkg(**args)


@depends()
def build(options: FixtureOptions, _fixtures: Fixtures) -> Build:
    args = get_options(options.get("build", {}), machine="lighthouse", build_id="34")

    return Build(**args)


@depends()
def tmpdir(_options: FixtureOptions, _fixtures: Fixtures) -> FixtureContext[Path]:
    with tempfile.TemporaryDirectory() as tempdir:
        yield Path(tempdir)


@depends("tmpdir")
def environ(
    options: FixtureOptions, fixtures: Fixtures
) -> FixtureContext[dict[str, str]]:
    mock_environ = {
        "BUILD_PUBLISHER_API_KEY_ENABLE": "no",
        "BUILD_PUBLISHER_JENKINS_BASE_URL": "https://jenkins.invalid/",
        "BUILD_PUBLISHER_RECORDS_BACKEND": "memory",
        "BUILD_PUBLISHER_STORAGE_PATH": str(fixtures.tmpdir / "gbp"),
        "BUILD_PUBLISHER_WORKER_BACKEND": "sync",
        "BUILD_PUBLISHER_WORKER_THREAD_WAIT": "yes",
        "GBP_FL_RECORDS_BACKEND": "memory",
        **options.get("environ", {}),
    }
    with mock.patch.dict(os.environ, mock_environ):
        yield mock_environ


@depends()
def now(options: FixtureOptions, _fixtures: Fixtures) -> dt.datetime:
    time: dt.datetime = options.get(
        "now", dt.datetime(2025, 1, 26, 12, 57, 37, tzinfo=dt.UTC)
    )
    return time


@depends()
def console(_options: FixtureOptions, _fixtures: Fixtures) -> FixtureContext[Console]:
    """StringIO Console"""
    out = io.StringIO()
    err = io.StringIO()

    c = Console(
        out=rich.console.Console(
            file=out, width=88, theme=Theme(DEFAULT_THEME), highlight=False, record=True
        ),
        err=rich.console.Console(file=err, record=True),
    )
    yield c

    if "SAVE_VIRTUAL_CONSOLE" in os.environ:
        global COUNTER  # pylint: disable=global-statement

        COUNTER += 1
        filename = f"{COUNTER}.svg"
        c.out.save_svg(filename, title="gbp-fl")


################################
# gentoo-build-publisher stuff #
################################
@depends("gbp_settings")
def publisher(
    _options: FixtureOptions, fixtures: Fixtures
) -> FixtureContext[BuildPublisher]:
    bp: BuildPublisher = BuildPublisher.from_settings(fixtures.gbp_settings)
    names = ["storage", "jenkins", "repo"]
    contexts = (
        mock.patch.object(publisher_obj, name, getattr(bp, name)) for name in names
    )

    with ExitStack() as stack:
        for cm in contexts:
            stack.enter_context(cm)

        yield bp


@depends("gbp_settings")
def worker(
    _options: FixtureOptions, fixtures: Fixtures
) -> FixtureContext[gbp_worker.WorkerInterface]:
    sync_worker = gbp_worker.Worker(fixtures.gbp_settings)
    with mock.patch("gentoo_build_publisher.worker", sync_worker):
        yield sync_worker


@depends("environ")
def gbp_settings(_options: FixtureOptions, _fixtures: Fixtures) -> GBPSettings:
    return GBPSettings.from_environ()


@depends()
def build_record(options: FixtureOptions, _fixtures: Fixtures) -> BuildRecord:
    record = get_options(
        options.get("build_record", {}), build_id="1502", machine="babette", note=None
    )
    return BuildRecord(**record)


@depends("build_record", "now")
def gbp_package(options: FixtureOptions, fixtures: Fixtures) -> gbp.Package:
    pkg_options = get_options(
        options.get("gbp_package", {}),
        build_id=1,
        build_time=fixtures.now.timestamp(),
        cpv="sys-libs/mtdev-1.1.7",
        path="sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
        repo="gentoo",
        size=40960,
    )
    return gbp.Package(**pkg_options)


def gbp_client(options: FixtureOptions, _fixtures: Fixtures) -> GBP:
    url = options.get("gbp", {}).get("url", "http://gbp.invalid/")
    gbp_ = GBP(url)
    gbp_.query._session.mount(  # pylint: disable=protected-access
        url, DjangoToRequestsAdapter()
    )

    return gbp_


def get_options(options: FixtureOptions, **defaults: Any) -> FixtureOptions:
    return {item: options.get(item, default) for item, default in defaults.items()}


DEFAULT_CONTENTS = """
    lighthouse 34 app-shells/bash-5.2_p37-1 /bin/bash
    lighthouse 34 app-shells/bash-5.2_p37-1 /etc/skel
    polaris    26 app-arch/tar-1.35-1       /bin/gtar
    polaris    26 app-shells/bash-5.2_p37-1 /bin/bash
    polaris    26 app-shells/bash-5.2_p37-2 /bin/bash
    polaris    27 app-shells/bash-5.2_p37-1 /bin/bash
"""


def seq_get(seq: Sequence[Any], index: int, default: Any = None) -> Any:
    """Like dict.get, but for sequences"""
    try:
        return seq[index]
    except IndexError:
        return default


class DjangoToRequestsAdapter(BaseAdapter):  # pylint: disable=abstract-method
    """Requests Adapter to call Django views"""

    def send(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        request: PreparedRequest,
        stream: bool = False,
        timeout: None | float | tuple[float, float] | tuple[float, None] = None,
        verify: bool | str = True,
        cert: None | bytes | str | tuple[bytes | str, bytes | str] = None,
        proxies: Mapping[str, str] | None = None,
    ) -> Response:
        assert request.method is not None
        django_response = Client().generic(
            request.method,
            request.path_url,
            data=request.body,
            content_type=request.headers["Content-Type"],
            **request.headers,
        )

        requests_response = Response()
        requests_response.raw = io.BytesIO(django_response.content)
        requests_response.raw.seek(0)
        requests_response.status_code = django_response.status_code
        requests_response.headers = CaseInsensitiveDict(django_response.headers)
        requests_response.encoding = django_response.get("Content-Type", None)
        requests_response.url = str(request.url)
        requests_response.request = request

        return requests_response



"""Test utilities"""

import datetime as dt
import os
from contextlib import contextmanager
from tarfile import TarInfo
from typing import Any, Generator

from django.test.client import Client

from gbp_fl.types import Build, BuildLike, Package

# pylint: disable=missing-docstring

LOCAL_TIMEZONE = dt.timezone(dt.timedelta(days=-1, seconds=61200), "PDT")


class MockGBPGateway:
    """Not a real gateway"""

    def __init__(self) -> None:
        self.builds: list[BuildLike] = []
        self.packages: dict[Build, list[Package]] = {}
        self.contents: dict[tuple[Build, Package], list[TarInfo]] = {}
        self.machines: list[str] = []

    def get_packages(self, build: Build) -> list[Package]:
        try:
            return self.packages[build]
        except KeyError:
            raise LookupError(build) from None

    def get_package_contents(self, build: Build, package: Package) -> list[TarInfo]:
        try:
            return self.contents[build, package]
        except KeyError:
            raise LookupError(build, package) from None

    def list_machine_names(self) -> list[str]:
        return self.machines


def graphql(query: str, variables: dict[str, Any] | None = None) -> Any:
    """Execute GraphQL query on the Django test client.

    Return the parsed JSON response
    """
    client = Client()
    response = client.post(
        "/graphql",
        {"query": query, "variables": variables},
        content_type="application/json",
    )

    return response.json()


@contextmanager
def cd(path: str) -> Generator[None, None, None]:
    cwd = os.getcwd()

    os.chdir(path)
    yield
    os.chdir(cwd)



"""Tests for gbp-fl"""

from functools import wraps
from typing import Any, Callable, TypeAlias, TypeVar
from unittest import TestCase

T = TypeVar("T", bound=TestCase)
TestFunc: TypeAlias = Callable[[T, *tuple[Any]], None]
Param: TypeAlias = list[Any]
Params: TypeAlias = list[Param]


def parametrized(lists_of_args: Params) -> Callable[[TestFunc[T]], TestFunc[T]]:
    """Turn TestCase test method into parametrized test"""

    def dec(func: TestFunc[T]) -> TestFunc[T]:
        @wraps(func)
        def wrapper(self: T, *args: Any, **kwargs: Any) -> None:
            for list_of_args in lists_of_args:
                name = ",".join(str(i) for i in list_of_args)
                with self.subTest(name):
                    func(self, *args, *list_of_args, **kwargs)

        return wrapper

    return dec



# pylint: disable=missing-docstring
import logging
import shutil
from pathlib import Path
from unittest import mock

import unittest_fixtures as uf

from gbp_fl import gateway
from gbp_fl.records import Repo
from gbp_fl.types import Build

TESTDIR = Path(__file__).parent

logging.basicConfig(handlers=[logging.NullHandler()])


@uf.requires("publisher")
class GBPTestCase(uf.TestCase):
    pass


@uf.depends("gbp_package", "build_record")
def binpkg(_o: uf.FixtureOptions, f: uf.Fixtures) -> Path:
    gbp = gateway.GBPGateway()
    path = Path(gbp.get_full_package_path(f.build_record, f.gbp_package))
    path.parent.mkdir(parents=True)
    shutil.copy(TESTDIR / "assets/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar", path)

    with open(path.parents[2] / "Packages", "w", encoding="utf8") as meta:
        meta.write(PACKAGES)

    return path


@uf.requires("worker", "gbp_package", "settings", binpkg)
class PostPulledTests(GBPTestCase):
    def test(self) -> None:
        f = self.fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        with mock.patch("gbp_fl.package_utils.Repo.from_settings", return_value=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        content_files = list(
            repo.files.for_package(
                f.build_record.machine,
                f.build_record.build_id,
                f"{f.gbp_package.cpv}-{f.gbp_package.build_id}",
            )
        )
        self.assertEqual(len(content_files), 10)

    def test_with_empty_tar(self) -> None:
        f = self.fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        shutil.copy(TESTDIR / "assets/empty.tar", f.binpkg)

        with mock.patch("gbp_fl.package_utils.Repo.from_settings", return_value=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        content_files = list(
            repo.files.for_package(
                f.build_record.machine, f.build_record.build_id, f.gbp_package.cpv
            )
        )
        self.assertEqual(len(content_files), 0)

    def test_pull_with_no_package_manifest(self) -> None:
        f = self.fixtures
        repo: Repo = Repo.from_settings(f.settings)
        gbp = gateway.GBPGateway()
        _ = None

        package_manifest: Path = f.binpkg.parents[2] / "Packages"
        package_manifest.unlink()

        with mock.patch("gbp_fl.package_utils.Repo.from_settings", return_value=repo):
            gbp.emit_signal(
                "postpull", build=f.build_record, packages=_, gbp_metadata=_
            )

        content_files = list(
            repo.files.for_package(
                f.build_record.machine,
                f.build_record.build_id,
                f"{f.gbp_package.cpv}-{f.gbp_package.build_id}",
            )
        )
        self.assertEqual(len(content_files), 0)


@uf.requires("worker", "settings", "bulk_content_files")
class PostDeleteTests(GBPTestCase):
    def test(self) -> None:
        f = self.fixtures
        repo: Repo = Repo.from_settings(f.settings)
        repo.files.bulk_save(f.bulk_content_files)
        gbp = gateway.GBPGateway()

        self.assertEqual(repo.files.count(None, None, None), 6)

        with mock.patch("gbp_fl.package_utils.Repo.from_settings", return_value=repo):
            build = Build(machine="polaris", build_id="26")
            gbp.emit_signal("postdelete", build=build)

        self.assertEqual(repo.files.count(None, None, None), 3)


@uf.requires("build_record", "gbp_package", binpkg)
class GetPackageContentsTests(GBPTestCase):
    def test(self) -> None:
        f = self.fixtures
        gbp = gateway.GBPGateway()

        contents = gbp.get_package_contents(f.build_record, f.gbp_package)

        self.assertEqual(len(list(contents)), 19)


PACKAGES = """\
ACCEPT_KEYWORDS: amd64 ~amd64
ACCEPT_LICENSE: @FREE
ACCEPT_PROPERTIES: *
ACCEPT_RESTRICT: *
ARCH: amd64
CBUILD: x86_64-pc-linux-gnu
CHOST: x86_64-pc-linux-gnu
ELIBC: glibc
GENTOO_MIRRORS: http://distfiles.gentoo.org
INSTALL_MASK: /usr/share/doc
KERNEL: linux
PACKAGES: 1
TIMESTAMP: 1738033770
VERSION: 0

BUILD_ID: 1
BUILD_TIME: 1725728729
CPV: sys-libs/mtdev-1.1.7
DEFINED_PHASES: configure install
DEPEND: >=sys-kernel/linux-headers-2.6.31
EAPI: 8
KEYWORDS: ~alpha amd64 arm arm64 ~hppa ~ia64 ~loong ~m68k ~mips ppc ppc64 ~riscv ~s390 sparc x86
LICENSE: MIT
MD5: 5c9f95bf9b86a05e7348e14b64badd69
PATH: sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar
PROVIDES: x86_64: libmtdev.so.1
RDEPEND: >=sys-libs/glibc-2.40
REQUIRES: x86_64: libc.so.6
SHA1: 837d681085179a949527938c71f5f32d8fd8ffc9
SIZE: 40960
USE: abi_x86_64 amd64 elibc_glibc kernel_linux
MTIME: 1725728730
REPO: gentoo
"""



"""Django Models for gbp-fl"""

import os.path
from typing import Any

from django.db import models


class ContentFile(models.Model):
    """DB backend for gbp-fl ContentFiles"""

    machine = models.CharField(max_length=255)
    build_id = models.CharField(max_length=255)
    path = models.CharField(max_length=1023)
    basename = models.CharField(max_length=255, db_index=True)
    cpvb = models.CharField(max_length=255)
    repo = models.CharField(max_length=127)
    size = models.IntegerField()
    timestamp = models.DateTimeField()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                "machine", "build_id", "cpvb", "path", name="unique_path"
            )
        ]
        indexes = [
            models.Index("machine", "build_id", name="idx_build"),
            models.Index("machine", "build_id", "cpvb", name="idx_build_cpvb"),
        ]
        get_latest_by = ["timestamp"]
        ordering = ["machine", "timestamp"]

    def save(self, *args: Any, **kwargs: Any) -> None:
        self.basename = os.path.basename(self.path)

        super().save(*args, **kwargs)



"""Tests for the package_utils module"""

from pathlib import Path
from unittest import mock

import unittest_fixtures as uf

from gbp_fl import package_utils
from gbp_fl.records import files_backend
from gbp_fl.types import Build, ContentFileInfo

from .utils import MockGBPGateway

# pylint: disable=missing-docstring

MOCK_PREFIX = "gbp_fl.package_utils."


@uf.requires("bulk_packages")
class IndexPackagesTests(uf.TestCase):
    options = {
        "bulk_packages": """
            app-crypt/rhash-1.4.5
            dev-libs/libgcrypt-1.11.0-r2
            dev-libs/openssl-3.3.2-r2
            dev-libs/wayland-protocols-1.39
            net-dns/c-ares-1.34.4
        """
    }

    def test(self) -> None:
        mock_gw = MockGBPGateway()
        build = Build(machine="babette", build_id="1505")
        package = self.fixtures.bulk_packages[0]
        mock_gw.packages[build] = [package]
        mock_tarinfo = mock.Mock(mtime=0, size=22)
        mock_tarinfo.isdir.return_value = False
        mock_tarinfo.name = "image/bin/bash"
        mock_gw.contents[build, package] = [mock_tarinfo]
        repo = mock.Mock(files=files_backend("memory"))

        with mock.patch(f"{MOCK_PREFIX}GBPGateway", return_value=mock_gw):
            with mock.patch(f"{MOCK_PREFIX}Repo.from_settings", return_value=repo):
                package_utils.index_packages(build)

        self.assertEqual(repo.files.count(None, None, None), 1)

        content_file = next(iter(repo.files.files.values()))
        self.assertEqual(content_file.path, Path("/bin/bash"))
        self.assertEqual(content_file.size, 22)

    def test_when_no_package(self) -> None:
        mock_gw = MockGBPGateway()
        build = Build(machine="babette", build_id="1505")
        repo = mock.Mock(files=files_backend("memory"))

        with mock.patch(f"{MOCK_PREFIX}GBPGateway", return_value=mock_gw):
            with mock.patch(f"{MOCK_PREFIX}Repo.from_settings", return_value=repo):
                package_utils.index_packages(build)

        self.assertEqual(repo.files.count(None, None, None), 0)


@uf.requires("build_record", "gbp_package")
class MakeContentFileTests(uf.TestCase):
    def test(self) -> None:
        f = self.fixtures
        info = ContentFileInfo(name="/bin/bash", mtime=1738258812, size=8829)

        result = package_utils.make_content_file(f.build_record, f.gbp_package, info)

        self.assertEqual(result.path, Path("/bin/bash"))
        self.assertEqual(result.size, 8829)



"""Handler for `gbp fl fetch`"""

import argparse
import re
from dataclasses import dataclass

import requests
from gbpcli.gbp import GBP
from gbpcli.types import Console
from yarl import URL

# lighthouse/32284/www-client/firefox-135.0-1::gentoo
PKGSPEC_RE_STR = r"""
(?P<machine>[a-z]\w*)/
(?P<build_id>[0-9]+)/
(?P<c>[a-z0-9]+-[a-z0-9]+)/
(?P<p>[a-z].*)-(?P<v>[0-9].*)-(?P<b>[0-9]*)
"""

BUFSIZE = 1024
PKGSPEC_RE = re.compile(PKGSPEC_RE_STR, re.I | re.X)


@dataclass
class Parsed:
    """Parsed package spec"""

    machine: str
    build_id: str
    c: str
    p: str
    v: str
    b: int


def handler(args: argparse.Namespace, gbp: GBP, console: Console) -> int:
    """Get package files from Gentoo Build Publisher"""
    if (spec := parse_pkgspec(args.pkgspec)) is None:
        console.err.print(f"[red]Invalid specifier: {args.pkgspec}[/red]")
        return 1

    path = (
        f"machines/{spec.machine}/builds/{spec.build_id}/packages/{spec.c}"
        f"/{spec.p}/{spec.p}-{spec.v}-{spec.b}"
    )
    url = URL(gbp.query._url).origin() / path  # pylint: disable=protected-access

    with requests.get(str(url), stream=True, timeout=300) as response:
        if response.status_code == 404:
            console.err.print("[red]The requested package was not found.[/red]")
            return 2

        output = URL(response.url).name

        with open(output, "wb", buffering=BUFSIZE) as fp:
            for chunk in response.iter_content(BUFSIZE):
                fp.write(chunk)
        console.out.print(f"package saved as {output}")

    return 0


HELP = handler.__doc__


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Build command-line arguments"""
    parser.add_argument("pkgspec")


def parse_pkgspec(pkgspec: str) -> Parsed | None:
    """Parse the given spec"""
    if match := PKGSPEC_RE.match(pkgspec):
        parsed = Parsed(**match.groupdict())  # type: ignore
        parsed.b = int(parsed.b)
        return parsed
    return None



"""Tests for gbp-fl async tasks"""

# The tasks, by design, do basically nothing. We just have to assert the call the
# appropriate functions with the appropriate args

from unittest import mock

import unittest_fixtures as uf

from gbp_fl.types import Build
from gbp_fl.worker import tasks

# pylint: disable=missing-docstring


class IndexPackagesTests(uf.TestCase):
    @mock.patch("gbp_fl.package_utils")
    def test(self, package_utils: mock.Mock) -> None:
        tasks.index_packages("babette", "1505")

        package_utils.index_packages.assert_called_once_with(
            Build(machine="babette", build_id="1505")
        )


class DeleteFromBuildTests(uf.TestCase):
    @mock.patch("gbp_fl.records.Repo.from_settings")
    def test(self, repo_from_settings: mock.Mock) -> None:
        tasks.delete_from_build("babette", "1505")

        repo = repo_from_settings.return_value
        repo.files.delete_from_build.assert_called_once_with("babette", "1505")



"""Utilities for working with Packages"""

import datetime as dt
from concurrent.futures import ThreadPoolExecutor, wait
from pathlib import PurePath as Path

from gbp_fl.gateway import GBPGateway
from gbp_fl.records import Repo
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, ContentFile, ContentFileInfo, Package


def index_packages(build: Build) -> None:
    """Save the given Build's packages to the database"""
    gbp = GBPGateway()
    executor = ThreadPoolExecutor()

    try:
        packages = gbp.get_packages(build) or []
    except LookupError:
        return

    wait(executor.submit(index_package, package, build) for package in packages)


def index_package(package: Package, build: Build) -> None:
    """Save the files from the given build/package"""
    gbp = GBPGateway()
    package_contents = gbp.get_package_contents
    repo = Repo.from_settings(Settings.from_environ())

    content_files = (
        make_content_file(
            build,
            package,
            ContentFileInfo(name=item.name, mtime=int(item.mtime), size=item.size),
        )
        for item in package_contents(build, package)
        if not item.isdir() and item.name.startswith("image/")
    )
    repo.files.bulk_save(content_files)


def make_content_file(
    build: Build, gbp_package: Package, metadata: ContentFileInfo
) -> ContentFile:
    """Return a ContentFile given the parameters"""
    timestamp = dt.datetime.fromtimestamp(metadata.mtime, tz=dt.UTC)
    path = Path(metadata.name.removeprefix("image"))
    binpkg = BinPkg(
        build=Build(machine=build.machine, build_id=build.build_id),
        cpvb=f"{gbp_package.cpv}-{gbp_package.build_id}",
        repo=gbp_package.repo,
        build_time=dt.datetime.fromtimestamp(gbp_package.build_time, dt.UTC),
    )
    return ContentFile(
        path=path, binpkg=binpkg, size=metadata.size, timestamp=timestamp
    )



# pylint: disable=missing-docstring

import datetime as dt
import inspect
from functools import partial
from importlib import import_module
from pathlib import PurePath as Path

import unittest_fixtures as uf
from django.test import TestCase

from gbp_fl.records import ContentFiles, RecordNotFound, Repo, django_orm, files_backend
from gbp_fl.settings import Settings
from gbp_fl.types import ContentFile

from . import parametrized

now = partial(dt.datetime.now, tz=dt.UTC)

BACKENDS = [["memory"], ["django"]]


@uf.requires("content_file", "bulk_content_files")
class ContentFilesTests(uf.TestCase, TestCase):
    # pylint: disable=too-many-public-methods
    @parametrized(BACKENDS)
    def test_supports_protocol(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        cls = type(files)
        alias = f"{cls.__module__}:{cls.__name__}"

        for name, prop in ContentFiles.__dict__.items():
            if name.startswith("_"):
                continue

            if callable(prop):
                self.assertTrue(
                    hasattr(cls, name), f"{alias} should have method {name}"
                )
                self.assertTrue(callable(prop), f"{alias}.{name} should be callable")
                prot_sig = inspect.signature(prop)
                cls_sig = inspect.signature(getattr(cls, name))
                self.assertEqual(
                    prot_sig, cls_sig, f"{alias}.{name} has the wrong signature"
                )
            else:
                self.assertIs(type(prop), type(getattr(cls, name)))

    @parametrized(BACKENDS)
    def test_save(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        content_file: ContentFile = self.fixtures.content_file

        files.save(content_file)

        record = files.get(
            content_file.binpkg.build.machine,
            content_file.binpkg.build.build_id,
            content_file.binpkg.cpvb,
            content_file.path,
        )

        self.assertIsInstance(record, ContentFile)
        self.assertEqual(record, content_file)

    @parametrized(BACKENDS)
    def test_bulk_save(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        content_files = self.fixtures.bulk_content_files

        files.bulk_save(content_files)

        self.assertEqual(files.count(None, None, None), 6)

    @parametrized(BACKENDS)
    def test_get(self, backend_type: str) -> None:
        fixtures = self.fixtures
        content_file = fixtures.content_file
        files = files_backend(backend_type)

        files.save(content_file)

        binpkg = content_file.binpkg
        build = binpkg.build
        record = files.get(
            build.machine, build.build_id, binpkg.cpvb, content_file.path
        )

        self.assertEqual(record, content_file)

    @parametrized(BACKENDS)
    def test_get_not_found(self, backend_type: str) -> None:
        fixtures = self.fixtures
        content_file = fixtures.content_file
        files = files_backend(backend_type)

        binpkg = content_file.binpkg
        build = binpkg.build
        with self.assertRaises(RecordNotFound):
            files.get(build.machine, build.build_id, binpkg.cpvb, content_file.path)

    @parametrized(BACKENDS)
    def test_delete(self, backend_type: str) -> None:
        fixtures = self.fixtures
        content_file = fixtures.content_file
        files = files_backend(backend_type)
        files.save(content_file)

        files.delete(content_file)

        binpkg = content_file.binpkg
        build = binpkg.build
        with self.assertRaises(RecordNotFound):
            files.get(build.machine, build.build_id, binpkg.cpvb, content_file.path)

    @parametrized(BACKENDS)
    def test_delete_not_found(self, backend_type: str) -> None:
        fixtures = self.fixtures
        content_file = fixtures.content_file
        files = files_backend(backend_type)

        with self.assertRaises(RecordNotFound):
            files.delete(content_file)

    @parametrized(BACKENDS)
    def test_delete_from_build(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        files.delete_from_build(machine="polaris", build_id="26")

        self.assertEqual(files.count("polaris", "26", None), 0)
        self.assertEqual(files.count(None, None, None), 3)

    @parametrized(BACKENDS)
    def test_count(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        self.assertEqual(files.count("lighthouse", "34", None), 2)
        self.assertEqual(
            files.count("lighthouse", "34", "app-shells/bash-5.2_p37-1"), 2
        )
        self.assertEqual(files.count("lighthouse", "34", "bogus-cat/bogus-pkg-1"), 0)

        self.assertEqual(files.count("polaris", "26", None), 3)
        self.assertEqual(files.count("polaris", "27", None), 1)
        self.assertEqual(files.count("polaris", None, None), 4)

        self.assertEqual(files.count("babette", "226", None), 0)

    @parametrized(BACKENDS)
    def test_count_all(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        self.assertEqual(files.count(None, None, None), 6)

    @parametrized(BACKENDS)
    def test_count_cpv_without_build_id(self, backend_type: str) -> None:
        files = files_backend(backend_type)

        with self.assertRaises(ValueError):
            files.count("lighthouse", None, "app-shells/bash-5.2_p37")

    @parametrized(BACKENDS)
    def test_for_package(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkgs = files.for_package("lighthouse", "34", "app-shells/bash-5.2_p37-1")
        paths = {pkg.path for pkg in pkgs}
        self.assertEqual(paths, {Path("/bin/bash"), Path("/etc/skel")})

        pkgs = files.for_package("polaris", "26", "app-shells/bash-5.2_p37-1")
        paths = {pkg.path for pkg in pkgs}
        self.assertEqual(paths, {Path("/bin/bash")})

        pkgs = files.for_package("babette", "226", "app-shells/bash-5.2_p37-1")
        paths = {pkg.path for pkg in pkgs}
        self.assertEqual(paths, set())

    @parametrized(BACKENDS)
    def test_for_build(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkgs = files.for_build("lighthouse", "34")
        total = len(list(pkgs))
        self.assertEqual(total, 2)

        pkgs = files.for_build("polaris", "26")
        total = len(list(pkgs))
        self.assertEqual(total, 3)

        pkgs = files.for_build("polaris", "27")
        total = len(list(pkgs))
        self.assertEqual(total, 1)

        pkgs = files.for_build("babette", "226")
        total = len(list(pkgs))
        self.assertEqual(total, 0)

    @parametrized(BACKENDS)
    def test_for_machine(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkgs = files.for_machine("lighthouse")

        total = len(list(pkgs))
        self.assertEqual(total, 2)

        pkgs = files.for_machine("polaris")

        total = len(list(pkgs))
        self.assertEqual(total, 4)

    @parametrized(BACKENDS)
    def test_search_full_path(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkg_files = list(files.search("/bin/bash"))
        self.assertEqual(len(pkg_files), 4)

        pkg_files = list(files.search("bin/bash"))
        self.assertEqual(len(pkg_files), 4)

    @parametrized(BACKENDS)
    def test_search_basename(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkg_files = list(files.search("bash"))
        self.assertEqual(len(pkg_files), 4)

    @parametrized(BACKENDS)
    def test_search_wildcard(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkg_files = list(files.search("sk*"))
        self.assertEqual(len(pkg_files), 1)

        pkg_files = list(files.search("*"))
        self.assertEqual(len(pkg_files), 6)

        pkg_files = list(files.search("*a*"))
        self.assertEqual(len(pkg_files), 5)

        pkg_files = list(files.search("z*"))
        self.assertEqual(len(pkg_files), 0)

        pkg_files = list(files.search("*ash"))
        self.assertEqual(len(pkg_files), 4)

    @parametrized(BACKENDS)
    def test_search_with_empty_string(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkg_files = list(files.search(""))
        self.assertEqual(len(pkg_files), 0)

    @parametrized(BACKENDS)
    def test_search_machine(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        pkg_files = list(files.search("bash", machine="polaris"))
        self.assertEqual(len(pkg_files), 3)

    @parametrized(BACKENDS)
    def test_exists_false(self, backend_type: str) -> None:
        files = files_backend(backend_type)

        exists = files.exists(
            "lighthouse", "34", "app-shells/bash-5.2_p37", "/dev/null"
        )

        self.assertIs(exists, False)

    @parametrized(BACKENDS)
    def test_exists_true(self, backend_type: str) -> None:
        files = files_backend(backend_type)
        files.bulk_save(self.fixtures.bulk_content_files)

        exists = files.exists(
            "lighthouse", "34", "app-shells/bash-5.2_p37-1", "/bin/bash"
        )

        self.assertIs(exists, True)


class ContentFilesBackendTests(uf.TestCase):
    def test_gets_given_backend(self) -> None:
        memory = import_module("gbp_fl.records.memory")

        backend = files_backend("memory")
        self.assertIsInstance(backend, memory.ContentFiles)

    def test_when_backend_not_found(self) -> None:
        with self.assertRaises(LookupError):
            files_backend("bogus")


class RepoFromSettingsTests(uf.TestCase):
    def test(self) -> None:
        settings = Settings(RECORDS_BACKEND="django")
        repo = Repo.from_settings(settings)

        self.assertIsInstance(repo.files, django_orm.ContentFiles)



"""memory-based ContentFiles backend"""

import fnmatch
from dataclasses import replace
from pathlib import PurePath as Path
from typing import Any, Iterable

from gbp_fl.types import ContentFile

from . import RecordNotFound


class ContentFiles:
    """Memory-backed ContentFiles repo"""

    def __init__(self) -> None:
        # [machine, build_id, cpvb, path, package_build_id] = ContentFile
        self.files: dict[tuple[str, str, str, str], ContentFile] = {}

    def save(self, content_file: ContentFile, **fields: Any) -> ContentFile:
        """Save the given ContentFile with given updated fields

        Return the updated ContentFile
        """
        files = self.files
        new = replace(content_file, **fields)

        try:
            self.delete(content_file)
        except RecordNotFound:
            pass

        binpkg = new.binpkg
        build = binpkg.build
        files[build.machine, build.build_id, binpkg.cpvb, str(new.path)] = new

        return new

    def bulk_save(self, content_files: Iterable[ContentFile]) -> None:
        """Bulk save a list of ContentFiles"""
        for content_file in content_files:
            self.save(content_file)

    def get(
        self, machine: str, build_id: str, cpvb: str, path: str | Path
    ) -> ContentFile:
        """Return the ContentFile with the given properties

        If no ContentFile matches, raise RecordNotFound
        """
        path = str(path)
        files = self.files
        if record := files.get((machine, build_id, cpvb, path), None):
            return record

        raise RecordNotFound()

    def delete(self, content_file: ContentFile) -> None:
        """Delete the given ContentFile from the database

        Raise RecordNotFound if it doesn't exist in the database.
        """
        files = self.files

        binpkg = content_file.binpkg
        build = binpkg.build
        try:
            del files[
                build.machine, build.build_id, binpkg.cpvb, str(content_file.path)
            ]
        except KeyError:
            raise RecordNotFound() from None

    def delete_from_build(self, machine: str, build_id: str) -> None:
        """Delete all content files for the given build"""
        match = (machine, build_id)
        files = self.files
        keys = tuple(files)

        for key in keys:
            if key[:2] == match:
                del files[key]

    def exists(self, machine: str, build_id: str, cpvb: str, path: str | Path) -> bool:
        """Return true if a package file with matching criteria exists in the db"""
        try:
            self.get(machine, build_id, cpvb, path)
            return True
        except RecordNotFound:
            return False

    def count(self, machine: str | None, build_id: str | None, cpvb: str | None) -> int:
        """Return the number of package files exist with the given critiria

        When the following parameters are not None:

        - machine: the number of files for the given machine
        - machine and build_id: the number of files for the given build
        - machine, build_id, and cpvb: the number of files for the given build's package

        When all parameters are none, returns the total number of package files on GBP

        All other combinations raise ValueError
        """
        params = {"machine": machine, "build_id": build_id, "cpvb": cpvb}
        query: tuple[str, ...] = ()

        previous = ""
        for i, (field, value) in enumerate(params.items()):
            if value:
                if len(query) < i:
                    raise ValueError(f"Must supply {previous} if supplying {field}")
                query = (*query, value)
            previous = field

        item_count = len(query)
        return sum(1 for record in self.files if record[:item_count] == query)

    def for_package(
        self, machine: str, build_id: str, cpvb: str
    ) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build and cpvb"""
        files = self.files.copy()
        for key, content_file in files.items():
            if key[:3] == (machine, build_id, cpvb):
                yield content_file

    def for_build(self, machine: str, build_id: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build"""
        files = self.files.copy()
        for key, content_file in files.items():
            if key[:2] == (machine, build_id):
                yield content_file

    def for_machine(self, machine: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given machine"""
        files = self.files.copy()
        for key, content_file in files.items():
            if key[0] == machine:
                yield content_file

    def search(self, key: str, machine: str | None = None) -> Iterable[ContentFile]:
        """Search the database for package files

        If machine is provided, restrict the search to files belonging to the given
        machine.

        The simple search key works like the following:

            - A key without "*" or "/" characters searches an exact match on the file's
              base name. For example if the key is "bash" then it matches "/bin/bash"
              but not "/usr/bin/bashbug"

            - Keys containing at least one "/" are interpreted as exact path matches.
              For example the key "/bin/bash" matches files whose path is exactly
              "/bin/bash". If the key does not start with a forward slash then it is
              automatically prepended.

            - Keys with an asterisk either at the start and/or end of the key perform
              wildcard matches but only on the basename of the file. For example the key
              "b*" matches "/bin/bash" and "/usr/bin/bashbug" but not
              "/usr/share/baselayout/fstab". Keys with an asterisk in the middle depend
              on the backend and are not guaranteed to provide the expected matches.

            - A key that's the empty string ("") matches nothing.

            - A key that contains nothing bug asterisks (e.g. "*") depends on the
              backend and are not guaranteed to provide the expected matches.
        """
        if not key:
            return

        matcher = path_basename_checker

        if "/" in key:
            matcher = exact_match_checker
        elif "*" in key:
            matcher = glob_checker

        for content_file in self.files.values():
            if machine and content_file.binpkg.build.machine != machine:
                continue
            if matcher(content_file, key):
                yield content_file


def exact_match_checker(content_file: ContentFile, key: str) -> bool:
    """Return True if key matches the exact path for the given ContentFile

    Otherwise return False.
    """
    if key[0] != "/":
        key = f"/{key}"

    return str(content_file.path) == key


def path_basename_checker(content_file: ContentFile, key: str) -> bool:
    """Return True if key matches the exact basename for the given ContentFile

    Otherwise return False.
    """
    return content_file.path.name == key


def glob_checker(content_file: ContentFile, key: str) -> bool:
    """Return True if key is a glob match for the basename of the given ContentFile

    Otherwise return False.
    """
    basename = content_file.path.name

    return fnmatch.fnmatch(basename, key)



"""Async tasks for gbp-fl"""

# pylint: disable=import-outside-toplevel


def index_packages(machine: str, build_id: str) -> None:
    """Index packages for the given build"""
    import logging

    from gbp_fl import package_utils
    from gbp_fl.types import Build

    logger = logging.getLogger(__package__)
    build = Build(machine=machine, build_id=build_id)

    logger.info("Saving packages for %s.%s", machine, build_id)
    package_utils.index_packages(build)


def delete_from_build(machine: str, build_id: str) -> None:
    """Delete all the files from the given build"""
    from gbp_fl.records import Repo
    from gbp_fl.settings import Settings

    repo = Repo.from_settings(Settings.from_environ())
    files = repo.files

    files.delete_from_build(machine, build_id)



"""Tests for gbp_fl.types"""

# pylint: disable=missing-docstring
from unittest_fixtures import TestCase, requires

from gbp_fl.types import BinPkg, Build


@requires("now")
class BinPkgTests(TestCase):
    def test_cpv(self) -> None:
        fixtures = self.fixtures
        build = Build(machine="lighthouse", build_id="32267")
        binpkg = BinPkg(
            build=build,
            cpvb="x11-apps/xhost-1.0.10-3",
            repo="gentoo",
            build_time=fixtures.now,
        )
        self.assertEqual(binpkg.cpv, "x11-apps/xhost-1.0.10")

    def test_build_id(self) -> None:
        fixtures = self.fixtures
        build = Build(machine="lighthouse", build_id="32267")
        binpkg = BinPkg(
            build=build,
            cpvb="x11-apps/xhost-1.0.10-3",
            repo="gentoo",
            build_time=fixtures.now,
        )
        self.assertEqual(binpkg.build_id, 3)



"""Tests for the gateway interface"""

# pylint: disable=missing-docstring

from contextlib import ExitStack
from pathlib import PurePath as Path
from unittest import mock

import gentoo_build_publisher
from gentoo_build_publisher import types as gtype
from unittest_fixtures import FixtureContext
from unittest_fixtures import FixtureOptions as O
from unittest_fixtures import Fixtures as F
from unittest_fixtures import TestCase, requires

from gbp_fl import gateway as gw
from gbp_fl.types import Build, Package

TESTDIR = Path(__file__).parent


def mock_publisher(_o: O, _f: F) -> FixtureContext[dict[str, mock.Mock]]:
    mocks = {"storage": mock.Mock(), "jenkins": mock.Mock(), "repo": mock.Mock()}
    contexts = (
        mock.patch.object(gentoo_build_publisher.publisher, name, value)
        for name, value in mocks.items()
    )

    with ExitStack() as stack:
        for cm in contexts:
            stack.enter_context(cm)

        yield mocks


build = Build(machine="babette", build_id="1442")
package = Package(
    cpv="sys-libs/mtdev-1.1.7",
    repo="gentoo",
    build_id=1,
    build_time=0,
    path="sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
)


@requires(mock_publisher)
class GetFullPackagePathTests(TestCase):
    def test(self) -> None:
        mocks = self.fixtures.mock_publisher
        storage = mocks["storage"]

        storage.get_path.return_value = Path("/binpkgs/babette.1442")
        gbp = gw.GBPGateway()

        full_package_path = gbp.get_full_package_path(build, package)

        self.assertEqual(
            str(full_package_path),
            "/binpkgs/babette.1442/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
        )


@requires(mock_publisher)
class GetPackagesTests(TestCase):
    def test(self) -> None:
        mocks = self.fixtures.mock_publisher
        storage = mocks["storage"]

        gbp_packages = [
            gtype.Package(
                cpv="sys-libs/mtdev-1.1.7",
                build_id=1,
                repo="gentoo",
                path="/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
                size=200,
                build_time=0,
            )
        ]
        storage.get_packages.return_value = gbp_packages
        gbp = gw.GBPGateway()

        packages = gbp.get_packages(build)

        expected = [
            Package(
                cpv="sys-libs/mtdev-1.1.7",
                repo="gentoo",
                build_id=1,
                build_time=0,
                path="/sys-libs/mtdev/mtdev-1.1.7-1.gpkg.tar",
            )
        ]
        self.assertEqual(packages, expected)


@requires(mock_publisher)
class GetPackageContentsTests(TestCase):
    def test(self) -> None:
        mocks = self.fixtures.mock_publisher
        storage = mocks["storage"]
        storage.get_path.return_value = TESTDIR / "assets"

        gbp = gw.GBPGateway()
        result = list(gbp.get_package_contents(build, package))

        self.assertEqual(len(result), 19)

    def test_when_empty_tarfile(self) -> None:
        gbp = gw.GBPGateway()

        with mock.patch.object(
            gbp, "get_full_package_path", return_value=TESTDIR / "assets/empty.tar"
        ):
            result = list(gbp.get_package_contents(build, package))

        self.assertEqual(len(result), 0)


class ReceiveSignalTests(TestCase):
    def test(self) -> None:
        dispatcher = mock.Mock()
        receiver = mock.Mock()
        signal = "sos"

        gbp = gw.GBPGateway()

        with mock.patch("gentoo_build_publisher.signals.dispatcher", dispatcher):
            gbp.receive_signal(receiver, signal)

        dispatcher.bind.assert_called_once_with(sos=receiver)


class EmitSignalTests(TestCase):
    def test(self) -> None:
        dispatcher = mock.Mock()
        signal = "sos"

        gbp = gw.GBPGateway()

        with mock.patch("gentoo_build_publisher.signals.dispatcher", dispatcher):
            gbp.emit_signal(signal, buckle_my="shoe")

        dispatcher.emit.assert_called_once_with(signal, buckle_my="shoe")


class RunTaskTests(TestCase):
    def test(self) -> None:
        gbp = gw.GBPGateway()
        func = mock.Mock()

        with mock.patch("gentoo_build_publisher.worker") as worker:
            gbp.run_task(func, 1, 2, buckle_my="shoe")

        worker.run.assert_called_once_with(func, 1, 2, buckle_my="shoe")


class ListMachineNamesTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock) -> None:
        machines = ["babette", "lighthouse", "polaris"]
        publisher.repo.build_records.list_machines.return_value = machines

        gbp = gw.GBPGateway()

        self.assertEqual(gbp.list_machine_names(), machines)
        publisher.repo.build_records.list_machines.assert_called_once_with()


class GetBuildsForMachineTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock) -> None:
        builds = [Build(machine="babette", build_id=f"150{i}") for i in range(5)]
        publisher.repo.build_records.for_machine.return_value = iter(builds)

        gbp = gw.GBPGateway()

        self.assertEqual(list(gbp.get_builds_for_machine("babette")), builds)
        publisher.repo.build_records.for_machine.assert_called_once_with("babette")


class GetBuildRecordTests(TestCase):
    @mock.patch("gentoo_build_publisher.publisher")
    def test(self, publisher: mock.Mock) -> None:
        bdict = {"machine": "babette", "build_id": "1507"}
        build_record = mock.Mock(**bdict)
        publisher.repo.build_records.get.return_value = build_record
        gbp = gw.GBPGateway()

        result = gbp.get_build_record(Build(**bdict))

        self.assertEqual(result, build_record)
        publisher.repo.build_records.get.assert_called_once_with(gtype.Build(**bdict))



"""Tests for the `gbp fl fetch` sub-sub-command"""

# pylint: disable=missing-docstring
import argparse
from typing import Any
from unittest.mock import ANY, MagicMock, patch

from unittest_fixtures import TestCase, requires

from gbp_fl.cli import fetch

from .utils import cd


@requires("tmpdir", "gbp_client", "repo", "console")
@patch("gbp_fl.cli.fetch.requests")
class HandlerTests(TestCase):
    def test(self, requests: MagicMock) -> None:
        args = argparse.Namespace(pkgspec="lighthouse/34/app-shells/bash-5.2_p37-1")
        gbp = self.fixtures.gbp_client
        console = self.fixtures.console
        pkg = "bash-5.2_p37-1"

        # This is a redirect that we let requests handle for us, so the requested URL
        # and the response URL will be different
        request_url = (
            "http://gbp.invalid/machines/lighthouse/builds/34/packages"
            f"/app-shells/bash/{pkg}"
        )
        response_url = (
            f"http://gbp.invalid/binpkgs/lighthouse.34/app-shells/bash/{pkg}.gpkg.tar"
        )

        mock_response = get_mock_response(url=response_url)

        requests.get.return_value = mock_response

        console.out.print(f"$ gbp fl fetch {args.pkgspec}")
        with cd(self.fixtures.tmpdir):
            status = fetch.handler(args, gbp, console)

        self.assertEqual(0, status)
        self.assertEqual("", console.err.file.getvalue())
        self.assertEqual(
            f"$ gbp fl fetch {args.pkgspec}\n" f"package saved as {pkg}.gpkg.tar\n",
            console.out.file.getvalue(),
        )

        with open(self.fixtures.tmpdir / f"{pkg}.gpkg.tar", "rb") as fp:
            content = fp.read()
        self.assertEqual(content, b"no matter")

        requests.get.assert_called_once_with(request_url, stream=True, timeout=ANY)

    def test_invalid_spec(self, requests: MagicMock) -> None:
        pkgspec = "lighthouse/34/bash-5.2_p37-1"
        args = argparse.Namespace(pkgspec=pkgspec)
        gbp = MagicMock()
        console = self.fixtures.console

        with cd(self.fixtures.tmpdir):
            status = fetch.handler(args, gbp, console)

        self.assertEqual(status, 1)
        self.assertEqual(f"Invalid specifier: {pkgspec}\n", console.err.file.getvalue())
        self.assertEqual("", console.out.file.getvalue())

        requests.get.assert_not_called()

    def test_when_server_returns_404(self, requests: MagicMock) -> None:
        pkgspec = "lighthouse/34/app-shells/bash-5.2_p37-1"
        args = argparse.Namespace(pkgspec=pkgspec)
        url = (
            "http://gbp.invalid/machines/lighthouse/builds/34/packages"
            "/app-shells/bash/bash-5.2_p37-1"
        )
        mock_response = get_mock_response(404, b"Not Found", url=url)
        gbp = self.fixtures.gbp_client
        console = self.fixtures.console
        requests.get.return_value = mock_response

        with cd(self.fixtures.tmpdir):
            status = fetch.handler(args, gbp, console)

        self.assertEqual(2, status)
        self.assertEqual(
            "The requested package was not found.\n", console.err.file.getvalue()
        )


class ParseArgsTests(TestCase):
    def test(self) -> None:
        parser = argparse.ArgumentParser()
        fetch.parse_args(parser)


def get_mock_response(
    status_code: int = 200, content: bytes = b"no matter", **attrs: Any
) -> MagicMock:
    mock_response = MagicMock(status_code=status_code)
    mock_response.iter_content.return_value = iter([content])
    for name, value in attrs.items():
        setattr(mock_response, name, value)

    mock_ctx = MagicMock()
    mock_ctx.__enter__.return_value = mock_response

    return mock_ctx



"""GraphQL interface for gbp-fl"""

# pylint: disable=missing-docstring

import re
from importlib import resources
from typing import Any

from ariadne import ObjectType, convert_kwargs_to_snake_case, gql
from django.http import HttpRequest
from graphql import GraphQLResolveInfo

from gbp_fl.gateway import GBPGateway
from gbp_fl.records import Repo
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, BuildLike, ContentFile

# Version regex for cpv's
V_RE = re.compile("-[0-9]")

type_defs = gql(resources.read_text("gbp_fl", "schema.graphql"))
resolvers = [
    binpkg := ObjectType("flBinPkg"),
    content_file := ObjectType("flContentFile"),
    query := ObjectType("Query"),
    mutation := ObjectType("Mutation"),
]


@query.field("flSearch")
def resolve_file_list_search(
    _obj: Any, _info: GraphQLResolveInfo, *, key: str, machine: str | None = None
) -> list[ContentFile]:
    files = get_repo().files

    return list(files.search(key, machine))


@query.field("flCount")
@convert_kwargs_to_snake_case
def resolve_file_list_count(
    _obj: Any,
    _info: GraphQLResolveInfo,
    *,
    machine: str | None = None,
    build_id: str | None = None,
) -> int:
    files = get_repo().files

    return files.count(machine, build_id, None)


@binpkg.field("build")
def resolve_binpkg_build(pkg: BinPkg, _info: GraphQLResolveInfo) -> BuildLike:
    build = pkg.build
    gbp = GBPGateway()

    return gbp.get_build_record(Build(machine=build.machine, build_id=build.build_id))


@binpkg.field("url")
def resolve_binpkg_url(pkg: BinPkg, info: GraphQLResolveInfo) -> str:
    cpv = pkg.cpv
    c, pv = cpv.split("/", 1)

    v_match = V_RE.search(pv)
    assert v_match is not None
    p = pv[: v_match.start()]
    request: HttpRequest = info.context["request"]
    build = pkg.build

    return request.build_absolute_uri(
        f"/machines/{build.machine}/builds/{build.build_id}/packages/{c}/{p}"
        f"/{pv}-{pkg.build_id}"
    )


def get_repo() -> Repo:
    return Repo.from_settings(Settings.from_environ())



"""DB interface for gbp-fl"""

import importlib.metadata
from dataclasses import dataclass
from pathlib import PurePath as Path
from typing import Any, Iterable, Protocol, Self, cast

from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, ContentFile


class RecordNotFound(LookupError):
    """Exception raised when a requested record was not found"""


class ContentFiles(Protocol):  # pragma: no cover
    """Repository for Package files"""

    def save(self, content_file: ContentFile, **fields: Any) -> ContentFile:
        """Save the given ContentFile with given updated fields

        Return the updated ContentFile
        """

    def bulk_save(self, content_files: Iterable[ContentFile]) -> None:
        """Bulk save a list of ContentFiles"""

    def get(
        self, machine: str, build_id: str, cpvb: str, path: str | Path
    ) -> ContentFile:
        """Return the ContentFile with the given properties

        If no ContentFile matches, raise RecordNotFound
        """

    def delete(self, content_file: ContentFile) -> None:
        """Delete the given ContentFile from the database

        Raise RecordNotFound if it doesn't exist in the database.
        """

    def delete_from_build(self, machine: str, build_id: str) -> None:
        """Delete all content files for the given build"""

    def exists(self, machine: str, build_id: str, cpvb: str, path: str | Path) -> bool:
        """Return true if a package file with matching criteria exists in the db"""

    def count(self, machine: str | None, build_id: str | None, cpvb: str | None) -> int:
        """Return the number of package files exist with the given critiria

        When the following parameters are not None:

        - machine: the number of files for the given machine
        - machine and build_id: the number of files for the given build
        - machine, build_id, and cpv: the number of files for the given build's package

        When all parameters are none, returns the total number of package files on GBP

        All other combinations raise ValueError
        """

    def for_package(
        self, machine: str, build_id: str, cpvb: str
    ) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build and cpv"""

    def for_build(self, machine: str, build_id: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build"""

    def for_machine(self, machine: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given machine"""

    def search(self, key: str, machine: str | None = None) -> Iterable[ContentFile]:
        """Search the database for package files

        If machine is provided, restrict the search to files belonging to the given
        machine.

        The simple search key works like the following:

            - A key without "*" or "/" characters searches an exact match on the file's
              base name. For example if the key is "bash" then it matches "/bin/bash"
              but not "/usr/bin/bashbug"

            - Keys containing at least one "/" are interpreted as exact path matches.
              For example the key "/bin/bash" matches files whose path is exactly
              "/bin/bash". If the key does not start with a forward slash then it is
              automatically prepended.

            - Keys with an asterisk either at the start and/or end of the key perform
              wildcard matches but only on the basename of the file. For example the key
              "b*" matches "/bin/bash" and "/usr/bin/bashbug" but not
              "/usr/share/baselayout/fstab". Keys with an asterisk in the middle depend
              on the backend and are not guaranteed to provide the expected matches.

            - A key that's the empty string ("") matches nothing.

            - A key that contains nothing bug asterisks (e.g. "*") depends on the
              backend and are not guaranteed to provide the expected matches.
        """


def files_backend(backend: str) -> ContentFiles:
    """Load the ContentFiles db interface given the settings"""
    try:
        [module] = importlib.metadata.entry_points(group="gbp_fl.records", name=backend)
    except ValueError:
        raise LookupError(f"RECORDS_BACKEND not found: {backend}") from None

    return cast(ContentFiles, module.load().ContentFiles())


@dataclass(frozen=True)
class Repo:
    """Repository pattern"""

    files: ContentFiles

    @classmethod
    def from_settings(cls: type[Self], settings: Settings) -> Self:
        """Return instance of the Repo class given in settings"""
        return cls(files=files_backend(settings.RECORDS_BACKEND))



# pylint: disable=missing-docstring
import argparse
from unittest import mock

from unittest_fixtures import TestCase, requires

from gbp_fl import cli


@requires("console")
class HandlerTests(TestCase):
    def test(self) -> None:
        args = argparse.Namespace()
        gbp = mock.Mock()
        console = self.fixtures.console

        status = cli.handler(args, gbp, console)

        self.assertEqual(status, 1)
        self.assertEqual(console.out.file.getvalue(), "")
        self.assertTrue(console.err.file.getvalue().startswith("Subcommands:"))


class ParseArgsTests(TestCase):
    def test(self) -> None:
        parser = argparse.ArgumentParser()
        cli.parse_args(parser)



"""AppsConfig for gbp-fl"""

from importlib import import_module

from django.apps import AppConfig


class GBPFLConfig(AppConfig):
    """AppConfig for gbp-fl"""

    name = "gbp_fl"
    verbose_name = "File Lists for Gentoo Build Publisher"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self) -> None:
        signals = import_module("gbp_fl.signals")
        signals.init()



"""Settings for gbp-fl"""

from dataclasses import dataclass

from gbpcli.settings import BaseSettings


@dataclass(frozen=True)
class Settings(BaseSettings):
    """gbp-fl Settings"""

    env_prefix = "GBP_FL_"

    # pylint: disable=invalid-name
    RECORDS_BACKEND: str = "django"
    RECORDS_BACKEND_DJANGO_BULK_BATCH_SIZE: int = 300



"""gbp-fl's command-line interface

The gbpcli handler does nothing but print usage as `gbp fl` has subparsers itself and
there is nothing (yet) to do at the "root" level.
"""

import argparse
import functools
import importlib
from typing import IO

from gbpcli.gbp import GBP
from gbpcli.types import Console

HELP = "Access the Gentoo Build Publisher File List"
SUBCOMMANDS = ("search", "fetch")


def handler(args: argparse.Namespace, _gbp: GBP, console: Console) -> int:
    """GBP File List"""
    print_usage(console.err.file)
    return 1


handler.__doc__ = HELP


def parse_args(parser: argparse.ArgumentParser) -> None:
    """Create parsers for subcommands"""
    subparsers = parser.add_subparsers()

    for subcommand in SUBCOMMANDS:

        module = importlib.import_module(f"gbp_fl.cli.{subcommand}")
        subparser = subparsers.add_parser(subcommand, help=module.HELP)
        module.parse_args(subparser)
        subparser.set_defaults(func=getattr(module, "handler"))


def print_usage(io: IO[str]) -> None:
    """Print the usage to the given file handle"""
    uprint = functools.partial(print, file=io)
    uprint("Subcommands:")

    for subcommand in SUBCOMMANDS:
        uprint(f"  {subcommand}")



# pylint: disable=missing-docstring
import datetime as dt
from argparse import ArgumentParser, Namespace
from dataclasses import replace
from unittest.mock import Mock, patch

from unittest_fixtures import TestCase, requires

from gbp_fl.cli import search

from .utils import LOCAL_TIMEZONE

DAY = dt.timedelta(days=1, minutes=11, seconds=12)


@requires("gbp_client", "repo", "bulk_content_files", "console")
@patch("gbp_fl.graphql.GBPGateway")
@patch("gbpcli.render.LOCAL_TIMEZONE", new=LOCAL_TIMEZONE)
class SearchTests(TestCase):
    def test(self, gpbgateway: Mock) -> None:
        cfs = self.fixtures.bulk_content_files
        repo = self.fixtures.repo
        now = self.fixtures.now

        for cf in cfs:
            cf = replace(cf, timestamp=now)
            repo.files.save(cf)
            now = now + DAY

        bash_file_indexes = [0, 3, 4, 5]

        gpbgateway.return_value.get_build_record.side_effect = tuple(
            Mock(
                machine=cfs[i].binpkg.build.machine,
                build_id=cfs[i].binpkg.build.build_id,
                id=cfs[i].binpkg.build.build_id,
            )
            for i in bash_file_indexes
        )
        args = Namespace(key="bash", machine=None)
        gbp = self.fixtures.gbp_client
        console = self.fixtures.console

        console.out.print("$ gbp fl search bash")
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST1_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )

    def test_with_missing_metadata(self, gpbgateway: Mock) -> None:
        # When a package is being deleted, e.g. during a purge run, the binpkg metadata
        # may be gone. This results in an error in the graphql call and the binpkg field
        # being set to null. We want to (silently) ignore those
        cfs = self.fixtures.bulk_content_files
        repo = self.fixtures.repo
        repo.files.bulk_save(cfs)

        bash_file_indexes = [0, 3, 4, 5]
        build_records = [
            Mock(
                machine=cfs[i].binpkg.build.machine,
                build_id=cfs[i].binpkg.build.build_id,
                id=cfs[i].binpkg.build.build_id,
            )
            for i in bash_file_indexes
        ]
        # make the second one return None
        build_records[1] = None  # type: ignore
        gpbgateway.return_value.get_build_record.side_effect = tuple(build_records)

        args = Namespace(key="bash", machine=None)
        gbp = self.fixtures.gbp_client
        console = self.fixtures.console

        console.out.print("$ gbp fl search bash")
        status = search.handler(args, gbp, console)

        self.assertEqual(status, 0)
        self.assertEqual(
            TEST2_SEARCH_OUTPUT,
            console.out.file.getvalue(),
            "\n" + console.out.file.getvalue(),
        )


class ParseArgsTests(TestCase):
    def test(self) -> None:
        parser = ArgumentParser()
        search.parse_args(parser)


TEST1_SEARCH_OUTPUT = """$ gbp fl search bash
╭───────┬───────────────────┬─────────────────────────────────────────┬──────────╮
│  Size │ Timestamp         │ Package                                 │ Path     │
├───────┼───────────────────┼─────────────────────────────────────────┼──────────┤
│850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash│
│850648 │ 01/29/25 06:31:13 │ polaris/26/app-shells/bash-5.2_p37-1    │ /bin/bash│
│850648 │ 01/30/25 06:42:25 │ polaris/26/app-shells/bash-5.2_p37-2    │ /bin/bash│
│850648 │ 01/31/25 06:53:37 │ polaris/27/app-shells/bash-5.2_p37-1    │ /bin/bash│
╰───────┴───────────────────┴─────────────────────────────────────────┴──────────╯
"""
TEST2_SEARCH_OUTPUT = """$ gbp fl search bash
╭───────┬───────────────────┬─────────────────────────────────────────┬──────────╮
│  Size │ Timestamp         │ Package                                 │ Path     │
├───────┼───────────────────┼─────────────────────────────────────────┼──────────┤
│850648 │ 01/26/25 05:57:37 │ lighthouse/34/app-shells/bash-5.2_p37-1 │ /bin/bash│
│850648 │ 01/26/25 05:57:37 │ polaris/26/app-shells/bash-5.2_p37-2    │ /bin/bash│
│850648 │ 01/26/25 05:57:37 │ polaris/27/app-shells/bash-5.2_p37-1    │ /bin/bash│
╰───────┴───────────────────┴─────────────────────────────────────────┴──────────╯
"""



# Generated by Django 5.1.5 on 2025-02-01 20:33

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ContentFile",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("machine", models.CharField(max_length=255)),
                ("build_id", models.CharField(max_length=255)),
                ("path", models.CharField(max_length=1023)),
                ("basename", models.CharField(db_index=True, max_length=255)),
                ("cpvb", models.CharField(max_length=255)),
                ("repo", models.CharField(max_length=127)),
                ("size", models.IntegerField()),
                ("timestamp", models.DateTimeField()),
            ],
            options={
                "ordering": ["machine", "timestamp"],
                "get_latest_by": ["timestamp"],
                "indexes": [
                    models.Index(
                        models.F("machine"), models.F("build_id"), name="idx_build"
                    ),
                    models.Index(
                        models.F("machine"),
                        models.F("build_id"),
                        models.F("cpvb"),
                        name="idx_build_cpvb",
                    ),
                ],
                "constraints": [
                    models.UniqueConstraint(
                        models.F("machine"),
                        models.F("build_id"),
                        models.F("cpvb"),
                        models.F("path"),
                        name="unique_path",
                    )
                ],
            },
        )
    ]



"""Django ORM-backed records backend"""

import os.path
from dataclasses import replace
from pathlib import PurePath as Path
from typing import Any, Iterable

from django.db import transaction

from gbp_fl import models
from gbp_fl.records import RecordNotFound
from gbp_fl.settings import Settings
from gbp_fl.types import BinPkg, Build, ContentFile

BULK_BATCH_SIZE = 100


class ContentFiles:
    """Django ORM-backed repository for Package files"""

    def save(self, content_file: ContentFile, **fields: Any) -> ContentFile:
        """Save the given ContentFile with given updated fields

        Return the updated ContentFile
        """
        new = replace(content_file, **fields)
        model = content_file_to_model(new)

        try:
            self.delete(content_file)
        except RecordNotFound:
            pass

        model.save()

        return new

    @transaction.atomic()
    def bulk_save(self, content_files: Iterable[ContentFile]) -> None:
        """Bulk save a list of ContentFiles"""
        get_basename = os.path.basename
        manager = models.ContentFile.objects
        settings = Settings.from_environ()
        batch_size = settings.RECORDS_BACKEND_DJANGO_BULK_BATCH_SIZE

        items = manager.bulk_create(
            (content_file_to_model(cf) for cf in content_files), batch_size=batch_size
        )
        for item in items:
            item.basename = get_basename(item.path)
        manager.bulk_update(items, ["basename"], batch_size=batch_size)

    def get(
        self, machine: str, build_id: str, cpvb: str, path: str | Path
    ) -> ContentFile:
        """Return the ContentFile with the given properties

        If no ContentFile matches, raise RecordNotFound
        """
        model = get_model(machine, build_id, cpvb, path)
        return model_to_content_file(model)

    def delete(self, content_file: ContentFile) -> None:
        """Delete the given ContentFile from the database

        Raise RecordNotFound if it doesn't exist in the database.
        """
        model = get_model(
            content_file.binpkg.build.machine,
            content_file.binpkg.build.build_id,
            content_file.binpkg.cpvb,
            content_file.path,
        )
        model.delete()

    def delete_from_build(self, machine: str, build_id: str) -> None:
        """Delete all content files for the given build"""
        manager = models.ContentFile.objects

        manager.filter(machine=machine, build_id=build_id).delete()

    def exists(self, machine: str, build_id: str, cpvb: str, path: str | Path) -> bool:
        """Return true if a package file with matching criteria exists in the db"""
        manager = models.ContentFile.objects
        query = manager.filter(
            machine=machine, build_id=build_id, cpvb=cpvb, path=str(path)
        )
        return query.exists()

    def count(self, machine: str | None, build_id: str | None, cpvb: str | None) -> int:
        """Return the number of package files exist with the given critiria

        When the following parameters are not None:

        - machine: the number of files for the given machine
        - machine and build_id: the number of files for the given build
        - machine, build_id, and cpvb: the number of files for the given build's package

        When all parameters are none, returns the total number of package files on GBP

        All other combinations raise ValueError
        """
        params = {"machine": machine, "build_id": build_id, "cpvb": cpvb}
        query_dict: dict[str, Any] = {}

        previous = ""
        for field, value in params.items():
            if value:
                if previous and previous not in query_dict:
                    raise ValueError(f"Must supply {previous} if supplying {field}")
                query_dict[field] = value
            previous = field

        return models.ContentFile.objects.filter(**query_dict).count()

    def for_package(
        self, machine: str, build_id: str, cpvb: str
    ) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build and cpvb"""
        query = models.ContentFile.objects.filter(
            machine=machine, build_id=build_id, cpvb=cpvb
        )
        yield from (model_to_content_file(model) for model in query)

    def for_build(self, machine: str, build_id: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given build"""
        query = models.ContentFile.objects.filter(machine=machine, build_id=build_id)

        yield from (model_to_content_file(model) for model in query)

    def for_machine(self, machine: str) -> Iterable[ContentFile]:
        """Return all ContentFiles for the given machine"""
        query = models.ContentFile.objects.filter(machine=machine)

        yield from (model_to_content_file(model) for model in query)

    def search(self, key: str, machine: str | None = None) -> Iterable[ContentFile]:
        """Search the database for package files

        If machine is provided, restrict the search to files belonging to the given
        machine.

        The simple search key works like the following:

            - A key without "*" or "/" characters searches an exact match on the file's
              base name. For example if the key is "bash" then it matches "/bin/bash"
              but not "/usr/bin/bashbug"

            - Keys containing at least one "/" are interpreted as exact path matches.
              For example the key "/bin/bash" matches files whose path is exactly
              "/bin/bash". If the key does not start with a forward slash then it is
              automatically prepended.

            - Keys with an asterisk either at the start and/or end of the key perform
              wildcard matches but only on the basename of the file. For example the key
              "b*" matches "/bin/bash" and "/usr/bin/bashbug" but not
              "/usr/share/baselayout/fstab". Keys with an asterisk in the middle depend
              on the backend and are not guaranteed to provide the expected matches.

            - A key that's the empty string ("") matches nothing.

            - A key that contains nothing bug asterisks (e.g. "*") depends on the
              backend and are not guaranteed to provided the expected matches.
        """
        if not key:
            return

        query_dict = {"machine": machine} if machine is not None else {}

        if "/" in key:
            if key[0] != "/":
                key = f"/{key}"
            query_dict["path"] = key
        elif key.startswith("*") and key.endswith("*"):
            query_dict["basename__contains"] = key.strip("*")
        elif key.startswith("*"):
            query_dict["basename__endswith"] = key.lstrip("*")
        elif key.endswith("*"):
            query_dict["basename__startswith"] = key.rstrip("*")
        else:
            query_dict["basename"] = key

        query = models.ContentFile.objects.filter(**query_dict)

        yield from (model_to_content_file(model) for model in query)


def get_model(
    machine: str, build_id: str, cpvb: str, path: Path | str
) -> models.ContentFile:
    """Return the Django model matching the given parameters

    If no model exists, raise RecordNotFound.
    """
    try:
        return models.ContentFile.objects.get(
            machine=machine, build_id=build_id, cpvb=cpvb, path=str(path)
        )
    except models.ContentFile.DoesNotExist:
        raise RecordNotFound from None


def content_file_to_model(content_file: ContentFile) -> models.ContentFile:
    """Convert the given ContentFile to a ContentFile Django model

    The model returned unsaved.
    """
    model = models.ContentFile()
    model.machine = content_file.binpkg.build.machine
    model.build_id = content_file.binpkg.build.build_id
    model.path = str(content_file.path)
    model.cpvb = content_file.binpkg.cpvb
    model.repo = content_file.binpkg.repo
    model.size = content_file.size
    model.timestamp = content_file.timestamp

    return model


def model_to_content_file(model: models.ContentFile) -> ContentFile:
    """Convert the given ContentFile Django model to the ContentFile dataclass"""
    build = Build(machine=model.machine, build_id=model.build_id)
    binpkg = BinPkg(
        cpvb=model.cpvb, build=build, build_time=model.timestamp, repo=model.repo
    )
    content_file = ContentFile(
        path=Path(model.path), binpkg=binpkg, timestamp=model.timestamp, size=model.size
    )
    return content_file
